# Generated by Django 4.2.6 on 2023-10-24 15:57

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbmailing", "0001_initial_squashed_squashed_0008_alter_mail_bcc_email_alter_mail_cc_email_and_more"),
        ("directory", "0005_entry_salutation"),
    ]

    operations = [
        migrations.DeleteModel(
            name="MailingSettings",
        ),
    ]
